.FixMap <- function(world, along, fun, thr, shift, fixAnt = FALSE) {
  # world must be a maps map
  # along must be 'lon', 'x' or 'lat', 'y'
  # fun must be `>` or `<`
  # thr must be the threshold to use with the comparison function
  # shift must be the shift to apply to the points beyond thr. If willing
  # to remove the parts of the continent that go beyond thr, shift must be
  # set to NA
  # fixAnt is a boolean that tells whether to add some points to the 
  # antarctic continent so that it can be filled properly

  # Auxiliar functions
  splitByNA <- function(x) {
    ids <- 1 + cumsum(is.na(x))
    not.na <- !is.na(x)
    split(x[not.na], ids[not.na])
  }

  mergeByNA <- function(x) {
    as.vector(unlist(sapply(x, function(y) c(y, NA))))
  }

  intersectThr <- function(a, b, thr, along) {
    v <- b - a
    m <- v[2] / v[1]
    n <- a[2] - m * a[1]
    if (along == 'x') {
      intersect <- m * thr + n
      c(thr, intersect)
    } else {
      intersect <- (thr - n) / m
      c(intersect, thr)
    }
  }

  # Parameter sanitization
  if (!is.character(along)) {
    stop("Parameter 'along' must be a character string.")
  }
  if (!along %in% c('lon', 'lat', 'x', 'y')) {
    stop("Parameter 'along' must be one of 'lat' or 'lon'.")
  }
  if (along == 'lon') {
    along <- 'x'
  } else if (along == 'lat') {
    along <- 'y'
  }

  # First we divide the map data in lands
  lands_x <- splitByNA(world$x)
  lands_y <- splitByNA(world$y)
  lands <- world$names
  if (length(lands) > length(lands_x)) {
    lands <- lands[1:length(lands_x)]
  } else {
    lands <- c(lands, rep("", length(lands_x) - length(lands)))
  }
  # Fix the antarctic if requested
  if (fixAnt) {
    ant_pos <- which.min(sapply(lands_y, min))
    #ant_pos <- which(lands == 'Antarctica')
    #lands_x[[ant_pos]][length(lands_x[[ant_pos]])] <- 180
    #lands_y[[ant_pos]][length(lands_y[[ant_pos]])] <- -90
    lands_x[[ant_pos]] <- c(-180, lands_x[[ant_pos]], 180)
    lands_y[[ant_pos]] <- c(-90, lands_y[[ant_pos]], -90)
  }
  new_land_n <- 1
  # For each land, we will check all the regions that go beyond the specified
  # threshold. For each overflowing region, we will add new points with the 
  # intersection at the specified threshold, and will shift it as specified.
  initial_lands <- length(lands)
  for (land in 1:initial_lands) {
    if (along == 'x') {
      bad <- fun(lands_x[[land]], thr)
    } else {
      bad <- fun(lands_y[[land]], thr)
    }
    bad_points <- which(bad)
    # If there are points beyond the threshold
    if (length(bad_points) > 0) {
      # If all are beyond the threshold we move the whole land
      if (length(lands_x[[land]]) - length(bad_points) == 0) {
        if (is.na(shift)) {
          lands_x[[land]] <- numeric(0)
          lands_y[[land]] <- numeric(0)
        } else {
          if (along == 'x') {
            lands_x[[land]] <- lands_x[[land]] + shift
          } else {
            lands_y[[land]] <- lands_y[[land]] + shift
          }
        }
      } else {
        # Reordering some points to make sure that the point sequence does not
        # start and end with points out of the threshold
        if (head(bad_points, 1) == 1 && tail(bad_points, 1) == length(lands_x[[land]])) {
          last_good <- tail(which(!bad), 1)
          lands_x[[land]] <- c(tail(lands_x[[land]], length(lands_x[[land]]) - last_good),
                               head(lands_x[[land]], last_good))
          lands_y[[land]] <- c(tail(lands_y[[land]], length(lands_y[[land]]) - last_good),
                               head(lands_y[[land]], last_good))
          bad_points <- bad_points[c(which(bad_points > last_good),
                                     which(bad_points <= last_good))]
        }
        points_to_replace <- c()
        replacement_points_x <- c()
        replacement_points_y <- c()
        new_land_x <- c()
        new_land_y <- c()
        first_portion_point <- 1
        current_point <- 1
        previous_point <- length(lands_x[[land]])
        previous_is_bad <- previous_point %in% bad_points
        if (bad_points[1] == current_point) {
          current_is_bad <- TRUE
        } else {
          current_is_bad <- FALSE
        }
        first_land_point <- c(lands_x[[land]][1], lands_y[[land]][1])
        #first_land_point_is_bad <- current_is_bad
        # If there are some points beyond the threshold, we iterate for each 
        # of them
        while (current_point <= length(lands_x[[land]])) {
          # If we see that there is an intersection with the threshold between
          # two consecutive points:
          if (previous_is_bad != current_is_bad) {
            # We calculate the intersection with the threshold
            intersect <- intersectThr(c(lands_x[[land]][previous_point],
                                        lands_y[[land]][previous_point]),
                                      c(lands_x[[land]][current_point],
                                        lands_y[[land]][current_point]),
                                      thr, along)
            # If the point where we detected the intersection is beyond the 
            # threshold, it means we start to parse a piece of land beyond
            # the threshold: we add the intersection as a new point for a new 
            # land piece, and we keep the intersection to be added to the 
            # rest of land below the threshold.
            if (current_is_bad) {
              new_land_x <- c(intersect[1])
              new_land_y <- c(intersect[2])
              first_portion_point <- current_point
              replacement_points_x <- c(intersect[1])
              replacement_points_y <- c(intersect[2])
            } else {
              # If the point where we detected the intersection is within the
              # threshold, it means we start to parse a piece of land that is
              # correct and that we finish and add the crop of land beyond the
              # threshold
              points_to_replace <- first_portion_point:(current_point - 1)
              new_land_x <- c(new_land_x, lands_x[[land]][points_to_replace], intersect[1])
              new_land_y <- c(new_land_y, lands_y[[land]][points_to_replace], intersect[2])
              if (!is.na(shift)) {
                lands_x[[initial_lands + new_land_n]] <- new_land_x + ifelse(along == 'x', shift, 0)
                lands_y[[initial_lands + new_land_n]] <- new_land_y + ifelse(along == 'y', shift, 0)
                lands <- c(lands, paste0(lands[land], '_', new_land_n))
                new_land_n <- new_land_n + 1
              }
              replacement_points_x <- c(replacement_points_x, intersect[1])
              replacement_points_y <- c(replacement_points_y, intersect[2])
              lands_x[[land]] <- c(head(lands_x[[land]], head(points_to_replace, 1) - 1),
                                   replacement_points_x,
                                   tail(lands_x[[land]], length(lands_x[[land]]) - tail(points_to_replace, 1)))
              lands_y[[land]] <- c(head(lands_y[[land]], head(points_to_replace, 1) - 1),
                                   replacement_points_y,
                                   tail(lands_y[[land]], length(lands_y[[land]]) - tail(points_to_replace, 1)))
              if (along == 'x') {
                bad_points <- which(fun(lands_x[[land]], thr))
              } else {
                bad_points <- which(fun(lands_y[[land]], thr))
              }
              current_point <- current_point - length(points_to_replace) + length(replacement_points_x)
              previous_point <- previous_point - length(points_to_replace) + length(replacement_points_x)
              points_to_replace <- c()
              replacement_points_x <- c()
              replacement_points_y <- c()
            }
          }
          previous_point <- current_point
          previous_is_bad <- current_is_bad
          current_is_bad <- (current_point + 1) %in% bad_points
          current_point <- current_point + 1
        }
        # Check and include the last row of bad points (if any)
        if (previous_is_bad) {
          #if (!first_land_point_is_bad) {
            intersect <- intersectThr(c(lands_x[[land]][previous_point],
                                        lands_y[[land]][previous_point]),
                                      c(first_land_point[1],
                                        first_land_point[2]),
                                      thr, along)
          #} else {
          #  intersect <- first_land_point
          #}
          points_to_replace <- first_portion_point:(current_point - 1)
          new_land_x <- c(new_land_x, lands_x[[land]][points_to_replace], intersect[1])
          new_land_y <- c(new_land_y, lands_y[[land]][points_to_replace], intersect[2])
          if (!is.na(shift)) {
            lands_x[[initial_lands + new_land_n]] <- new_land_x + ifelse(along == 'x', shift, 0)
            lands_y[[initial_lands + new_land_n]] <- new_land_y + ifelse(along == 'y', shift, 0)
            lands <- c(lands, paste0(lands[land], '_', new_land_n))
            new_land_n <- new_land_n + 1
          }
          replacement_points_x <- c(replacement_points_x, intersect[1])
          replacement_points_y <- c(replacement_points_y, intersect[2])
          lands_x[[land]] <- c(head(lands_x[[land]], head(points_to_replace, 1) - 1),
                               replacement_points_x,
                               tail(lands_x[[land]], length(lands_x[[land]]) - tail(points_to_replace, 1)))
          lands_y[[land]] <- c(head(lands_y[[land]], head(points_to_replace, 1) - 1),
                               replacement_points_y,
                               tail(lands_y[[land]], length(lands_y[[land]]) - tail(points_to_replace, 1)))
        }
      }
    }
  }

  world$x <- mergeByNA(lands_x)
  world$y <- mergeByNA(lands_y)
  world$names <- lands
  world$range <- c(min(world$x, na.rm = TRUE), max(world$x, na.rm = TRUE),
                   min(world$y, na.rm = TRUE), max(world$y, na.rm = TRUE))
  world
}

clim.colors <- function(n) {
  colorbar <- colorRampPalette(c("dodgerblue4", "dodgerblue1", "forestgreen", 
                                 "yellowgreen", "white", "white", "yellow", 
                                 "orange", "red", "saddlebrown"))
  colorbar(n)
}

# Thanks to Jealie and nesuribe, as seen in
# http://stackoverflow.com/questions/20977477/how-to-assign-a-specific-color-to-na-in-an-image-plot
image.nan <- function(x, y, z, zlim, col, na.color='gray', 
                      outside.below.color='black', 
                      outside.above.color='white', ...) {
  zstep <- (zlim[2] - zlim[1]) / length(col)  # step in the color palette
  newz.below.outside <- zlim[2] + 0.5 * zstep  # new z for values below zlim
  newz.above.outside <- zlim[2] + 1.5 * zstep  # new z for values above zlim
  newz.na <- zlim[2] + 2.5 * zstep  # new z for NA

  z[which(z < zlim[1])] <- newz.below.outside  # we affect newz.below.outside
  z[which(z > zlim[2])] <- newz.above.outside  # we affect newz.above.outside
  z[which(is.na(z > zlim[2]))] <- newz.na  # same for newz.na

  ##zlim[1] <- zlim[2] + zstep  # extend lower limit to include below value
  zlim[2] <- zlim[2] + 3 * zstep  # extend top limit to include the two new 
                                  # values above and na

  # we construct the new color range by including: na.color and na.outside
  col <- c(col, outside.below.color, outside.above.color, na.color)  

  image(x, y, z, zlim = zlim, col = col, ...) # we finally call image(...)
}

PlotEquiMap <- function(var, lon, lat, filled.continents = TRUE, 
                        drawleg = TRUE, center_lon = NULL, 
                        breaks = NULL, cols = NULL, varlim = NULL,
                        colNA = 'white', colSup = 'white', colInf = 'white',
                        map_pars = NULL, ...) {
  if (!(is.numeric(var) && (length(dim(var)) == 2))) {
    stop("Error: parameter 'var' should be a numeric matrix.")
  }
  if (!(is.numeric(lon) && (length(dim(lon)) <= 1)) || any(is.na(lon))) {
    stop("Error: parameter 'lon' should be a numeric vector without NA values.")
  }
  if (!(is.numeric(lat) && (length(dim(lat)) <= 1)) || any(is.na(lat))) {
    stop("Error: parameter 'lat' should be a numeric vector without NA values.")
  }
  if (!all(dim(var) == c(length(lat), length(lon)))) {
    stop("Error: parameters 'lon' and/or 'lat' don't match the size of 'var'.")
  }
  if (length(lon) < 2 || length(lat) < 2) {
    stop("Error: at least a 2 by 2 region should be provided.")
  }

  if (!(is.logical(drawleg) || (drawleg %in% c('E', 'S')))) {
    stop("Error: parameter 'drawleg' should be TRUE, FALSE, 'E' or 'S'.")
  }
  if (!is.null(center_lon)) {
    if (!is.numeric(center_lon)) {
      stop("Error: 'center_lon' must be NULL, TRUE or FALSE.")
    } else if (center_lon < -180 || center_lon > 180) {
      stop("Error: 'center_lon' must be in the range [-180, 180].")
    }
  }

  if (is.null(varlim)) {
    varlim <- c(min(var, na.rm = TRUE), max(var, na.rm = TRUE))
  } else {
    if (!is.numeric(varlim) || length(varlim) != 2) {
      stop("Error: 'varlim' should be a numeric vector of length 2.")
    }
  }
  if (!is.null(cols)) {
    if (!is.character(cols)) {
      stop("Error: 'cols' must be a character vector.")
    }
  }
  if (is.null(breaks)) {
    if (!is.null(cols)) {
      breaks <- length(cols) + 1
    } else {
      breaks <- 21
    }
  }
  if (is.numeric(breaks)) {
    if (length(breaks) == 1) {
      breaks <- seq(varlim[1], varlim[2], length.out = breaks)
    }
    if (is.null(cols)) {
      cols <- clim.colors(length(breaks) - 1)
    } else if (length(cols) != (length(breaks) - 1)) {
      stop("Error: inconsistent number of 'breaks' and 'cols'.")
    }
  } else {
    stop("Error: 'breaks' must be a numeric vector.")
  }

  default_map_pars <- list(interior = FALSE, fill = filled.continents, 
                           wrap = TRUE, resolution = 0)
  if (!is.null(map_pars)) {
    if (!is.list(map_pars)) {
      stop("Error: 'map' pars should be a list of parameters to the function 'maps::map'.")
    }
    default_map_pars[names(map_pars)] <- map_pars
  }
  map_pars <- default_map_pars

  # Keep the original device parameters
   old_par <- par(no.readonly = TRUE)

  # Reorder the longitudes and latitudes
  lon_reord <- sort(lon, index.return = TRUE)
  lon <- lon_reord$x
  lon_reord <- lon_reord$ix
  first_lon <- lon[1]
  last_lon <- tail(lon, 1)
  if (first_lon > 360 || last_lon > 360 || first_lon < -360 || last_lon < -360) {
    stop("Error: all 'lon' and 'lat' values must be in the range [-360, 360].")
  }
  if (last_lon - first_lon >= 360) {
    stop("Error: max(lon) - min(lon) must be smaller than 360 (i.e. there can't be overlapping values).")
  }
  lat_reord <- sort(lat, index.return = TRUE)
  lat <- lat_reord$x
  lat_reord <- lat_reord$ix
  first_lat <- lat[1]
  last_lat <- tail(lat, 1)
  if (first_lat > 90 || last_lat > 90 || first_lat < -90 || last_lat < -90) {
    stop("Error: all 'lon' and 'lat' values must be in the range [-360, 360].")
  }

  # Work out the best arrangement for legend
  requested_width <- last_lon - first_lon
  requested_height <- last_lat - first_lat
  display_width <- par('fin')[1]
  display_height <- par('fin')[2]
  if (drawleg != FALSE) {
    if ((requested_height/requested_width > display_height/requested_width) && 
        (drawleg != 'S')) {
      layout(matrix(1:2, 1, 2), widths = c(5, 1))
      drawleg <- 'E'
    } else {
      layout(matrix(1:2, 2, 1), heights = c(5, 1))
      drawleg <- 'S'
    }
  }

  # Detect automatically whether the data is across greenwich
  lon_gaps <- lon - c(tail(lon, 1) - 360, lon[-length(lon)])
  greenwich_gap <- lon_gaps[1]
  lon_space <- tail(lon_gaps, 1)
  if (greenwich_gap < min(lon_gaps[-1])) {
    print("Warning: irregular gap detected between last and first longitudes in 'lon'. If 'lon' is irregular, automatic detection of greenwich may be wrong. Use parameter 'greenwich' instead.")
  }
  if (((tail(lon, 1) - head(lon, 1)) / min(lon_gaps[-1])) > length(lon)) {
    across_green <- TRUE  
  } else {
    across_green <- FALSE
  }
  if (is.null(center_lon)) {
    center_lon <- ifelse(across_green, 0, 180)
  }
  center_lon <- lon[which.min(abs(lon - center_lon))[1]]
  right_lon <- center_lon + 180
  left_lon <- center_lon - 180
  # Arrange data in function of if plot across greenwich is requested or not
  # and select the corresponding map.
  #if (greenwich) {
  #  lon_reord <- c(lon_reord[which(lon >= 180)], 
  #                 lon_reord[which((lon >= -180) & (lon < 180))],
  #                 lon_reord[which(lon < -180)])
  #  lon <- c(lon[which(lon >= 180)], lon[which((lon >= -180) & (lon < 180))], 
  #           lon[which(lon < -180)])
  #  lon[which(lon >= 180)] <- lon[which(lon >= 180)] - 360
  #  lon[which(lon < -180)] <- lon[which(lon < -180)] + 360
  #} else {
  #  lon_reord <- c(lon_reord[which(lon >= 0)], lon_reord[which(lon < 0)])
  #  lon[which(lon < 0)] <- lon[which(lon < 0)] + 360
  #}
  lon_reord <- c(lon_reord[which(lon >= right_lon)],
                 lon_reord[which((lon >= left_lon) & (lon < right_lon))],
                 lon_reord[which(lon < left_lon)])
  lon[which(lon >= right_lon)] <- lon[which(lon >= right_lon)] - 360
  lon[which(lon < left_lon)] <- lon[which(lon < left_lon)] + 360
  map_to_draw <- 'world'

  # Reorder the input matrix
  var[, 1:dim(var)[2]] <- var[, lon_reord]
  var[1:dim(var)[1], ] <- var[lat_reord, ]
  # Plot the variable
  image.nan(as.vector(lon[lon_reord]), as.vector(lat), t(var), varlim, cols, 
            colNA, colInf, colSup, xlab = "", ylab = "", ...)
  # With GEOmap
  #data(coastmap, package = 'GEOmap', envir = environment())
  #coastmap$STROKES$col <- rep('black', length(coastmap$STROKES$col))
  #GEOmap::plotGEOmap(coastmap, MAPstyle = 2, add = TRUE)
  # With maps
  #map('world', orientation = c(90, 0, 0), wrap = TRUE)
  # First get the map data
  map_pars[c('database', 'add', 'plot', 'wrap')] <- list(database = map_to_draw, 
                                                         add = FALSE, plot = FALSE,
                                                         wrap = TRUE)
  world <- do.call(map, map_pars)
  # Then amend it as needed
  # If geenwich, if filled TRUE or FALSE in both cases, 
  # points beyond -180 and 180 have to be moved
  world <- .FixMap(world, 'lon', `>`, right_lon + (lon_space / 2), -360, filled.continents)
  world <- .FixMap(world, 'lon', `<`, left_lon - (lon_space / 2), 360, FALSE) 
  # Finally plot the map
  map_pars[c('database', 'add', 'plot')] <- list(database = world, 
                                                 add = TRUE, plot = TRUE)
  do.call(map, map_pars)

  # Draw the colorbar
  if (drawleg != FALSE) {
    if (drawleg == 'S') {
      image(1:length(cols), 1, t(t(1:length(cols))), axes = FALSE, col = cols,
            xlab = '', ylab = '')
    }
  }

  # Restore the original device parameters
  par(old_par)
}
